@@ -12,6 +12,7 @@ import pandas as pd
 from tqdm import tqdm
 from datetime import datetime
 from shap import TreeExplainer
+from shap import KernelExplainer
 
 from sklearn import metrics
 from sklearn.model_selection import KFold
@@ -110,9 +111,6 @@ class Pipeline:
         no_dirs [bool]: Whether to create files or not
         verbose [int]: Level of verbosity
         """
-        # Production initiation
-        self.bestModel = None
-        self.settings = None
         self.mainDir = 'AutoML/'
 
         # Copy arguments
@@ -174,10 +172,6 @@ class Pipeline:
         self.verbose = kwargs.get('verbose', 0)
         self.noDirs = kwargs.get('no_dirs', False)
 
-        # Monitoring
-        self._prediction_time = None
-        self._main_predictors = None
-
         # Checks
         assert self.mode in [None, 'regression', 'classification'], 'Supported modes: regression, classification.'
         assert 0 < self.informationThreshold < 1, 'Information threshold needs to be within [0, 1]'
@@ -190,6 +184,14 @@ class Pipeline:
         if self.includeOutput and not self.sequence:
             warnings.warn('[AutoML] IMPORTANT: strongly advices to not include output without sequencing.')
 
+        # Create dirs
+        if not self.noDirs:
+            self._create_dirs()
+            self._load_version()
+
+        # Store Pipeline Settings
+        self.settings = {'pipeline': kwargs, 'validation': {}, 'feature_set': ''}
+
         # Objective & Scorer
         self.scorer = None
         if self.objective is not None:
@@ -197,15 +199,6 @@ class Pipeline:
             assert self.objective in metrics.SCORERS.keys(), 'Metric not supported, look at sklearn.metrics'
             self.scorer = metrics.SCORERS[self.objective]
 
-        # Instance initiating
-        self.data = None
-        self.x = None
-        self.y = None
-        self.featureSets = None
-        self.results = None
-        self.n_classes = None
-        self.is_fitted = False
-
         # Required sub-classes
         self.dataSampler = DataSampler()
         self.dataProcesser = DataProcesser()
@@ -213,13 +206,19 @@ class Pipeline:
         self.featureProcesser = FeatureProcesser()
         self.driftDetector = DriftDetector()
 
-        # Create dirs
-        if not self.noDirs:
-            self._create_dirs()
-            self._load_version()
+        # Instance initiating
+        self.bestModel = None
+        self.data = None
+        self.x = None
+        self.y = None
+        self.featureSets = None
+        self.results = None
+        self.n_classes = None
+        self.is_fitted = False
 
-        # Store Pipeline Settings
-        self.settings = {'pipeline': kwargs, 'validation': {}, 'feature_set': ''}
+        # Monitoring
+        self._prediction_time = None
+        self._main_predictors = None
 
     # User Pointing Functions
     def get_settings(self) -> dict:
@@ -347,8 +346,6 @@ class Pipeline:
 
         # Process data
         x = self.dataProcesser.transform(x)
-        if x.astype('float32').replace([np.inf, -np.inf], np.nan).isna().sum().sum() != 0:
-            raise ValueError(f"Column(s) with NaN: {list(x.keys()[x.isna().sum() > 0])}")
 
         # Drift Check
         self.driftDetector.check(x)
@@ -371,6 +368,10 @@ class Pipeline:
         if self.standardize:
             x, y = self._transform_standardize(x, y)
 
+        # NaN test -- datetime should be taken care of by now
+        if x.astype('float32').replace([np.inf, -np.inf], np.nan).isna().sum().sum() != 0:
+            raise ValueError(f"Column(s) with NaN: {list(x.keys()[x.isna().sum() > 0])}")
+
         # Return
         return x, y
 
@@ -399,7 +400,7 @@ class Pipeline:
             predictions = self.bestModel.predict(x)
 
         # Stop timer
-        self._prediction_time = (time.time() - start_time) / len(x)
+        self._prediction_time = (time.time() - start_time) / len(x) * 1000
 
         # Calculate main predictors
         self._get_main_predictors(x)
@@ -431,7 +432,7 @@ class Pipeline:
         prediction = self.bestModel.predict_proba(x)
 
         # Stop timer
-        self._prediction_time = (time.time() - start_time) / len(x)
+        self._prediction_time = (time.time() - start_time) / len(x) * 1000
 
         # Calculate main predictors
         self._get_main_predictors(x)
@@ -1063,6 +1064,7 @@ class Pipeline:
                 print('[AutoML] Loading existing model file.')
 
         # Update pipeline settings
+        self.settings['version'] = self.version
         self.settings['pipeline']['verbose'] = 0
         self.settings['model'] = model  # The string
         self.settings['params'] = params
@@ -1282,25 +1284,30 @@ class Pipeline:
         Using Shapely Additive Explanations, this function calculates the main predictors for a given
         prediction and sets them into the class' memory.
         """
-        if type(self.bestModel).__name__ in ['SVC', 'RidgeClassifier', 'LinearRegression']:
-            return
-        elif type(self.bestModel).__module__[:5] == 'Amplo':
-            explainer = TreeExplainer(self.bestModel.model)
+        # Shap is not implemented for all models.
+        if type(self.bestModel).__name__ in ['SVC', 'BaggingClassifier', 'RidgeClassifier', 'LinearRegression', 'SVR',
+                                             'BaggingRegressor']:
+            features = self.settings['feature_processing']['featureImportance']['shap'][0]
+            values = self.settings['feature_processing']['featureImportance']['shap'][1]
+            self._main_predictors = {features[i]: values[i] for i in range(len(features))}
+
         else:
-            explainer = TreeExplainer(self.bestModel)
+            if type(self.bestModel).__module__[:5] == 'Amplo':
+                shap_values = np.array(TreeExplainer(self.bestModel.model).shap_values(data))
+            else:
+                shap_values = np.array(TreeExplainer(self.bestModel).shap_values(data))
 
-        # Get values
-        shap_values = np.array(explainer.shap_values(data))
+            # Shape them (for multiclass it outputs ndim=3, for binary/regression ndim=2)
+            if shap_values.ndim == 3:
+                shap_values = shap_values[1]
 
-        # Shape them (for multiclass it outputs ndim=3, for binary/regression ndim=2)
-        if shap_values.ndim == 3:
-            shap_values = shap_values[1]
+            # Take mean over samples
+            shap_values = np.mean(shap_values, axis=0)
 
-        # Take mean over samples
-        shap_values = np.mean(shap_values, axis=0)
+            # Sort them
+            inds = sorted(range(len(shap_values)), key=lambda x: -abs(shap_values[x]))
 
-        # Sort them
-        inds = sorted(range(len(shap_values)), key=lambda x: -abs(shap_values[x]))
+            # Set class attribute
+            self._main_predictors = dict([(data.keys()[i], float(abs(shap_values[i]))) for i in inds])
 
-        # Set class attribute
-        self._main_predictors = dict([(data.keys()[i], float(abs(shap_values[i]))) for i in inds])
+        return self._main_predictors
