@@ -10,7 +10,7 @@ class TestRegression(object):
         self.y = None
 
     def test_set_params(self):
-        self.model.set_params(**{'depth': 10})
+        self.model.set_params(**{'max_depth': 10})
 
     def test_get_params(self):
         self.model.get_params()
@@ -38,7 +38,7 @@ class TestRegression(object):
         assert len(prediction.shape) == 1
 
     def test_cloneable(self):
-        params = {'depth': 5}
+        params = {'max_depth': 5}
         cloned = clone(clone(self.model).set_params(**params))
         cloned.fit(self.x, self.y)
         prediction = cloned.predict(self.x)
